from typing import Set

from dataclasses import dataclass
from datetime import datetime


@dataclass
class Nav:
    """Class for store the NAV value with references"""

    value: float
    updated: datetime
    tags: Set[str]
    fund: str



import datetime

import dateparser


def date_range(start_date, end_date):
    return [
        start_date + datetime.timedelta(days=x)
        for x in range((end_date - start_date).days + 1)
    ]


def convert_buddhist_to_gregorian(input_date):
    if isinstance(input_date, str):
        input_date = dateparser.parse(input_date)
    year = input_date.year - 543
    input_date = input_date.replace(year=year)
    return input_date



from typing import List

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from . import sources
from .nav import Nav
from .utils._optional import import_optional_dependency


def get(fund_name, *, source="finnomena", date=None, **kargs) -> Nav:
    """
    Gets the latest NAV

    **Parameters:**

    * **fund_name** - Fund name found in finnomena such as `TISTECH-A`
    * **source** - *(optional)* Data source for pull data. See Data Sources
    section in the documentation for all availiable options.
    * **date** - *(optional)* get latest price of a given date
    * **subscription_key** - *(optional)* Subscription key that required for
    a data source like `sec` (a.k.a)

    **Returns:** `Nav`

    Usage:
    ```
    >>> import pythainav as nav

    >>> nav.get("KT-PRECIOUS")
    Nav(value=4.2696, updated='20/01/2020', tags={'latest'}, fund='KT-PRECIOUS')
    ```
    """
    fund_name = fund_name

    source2class = {
        "finnomena": sources.Finnomena,
        "sec": sources.Sec,
        # "onde": sources.Onde,
    }
    _source = source2class[source](**kargs)

    nav = _source.get(fund_name, date)

    return nav


def get_all(
    fund_name,
    *,
    source="finnomena",
    asDataFrame=False,
    range: Literal[
        "1D", "1W", "1M", "6M", "YTD", "1Y", "3Y", "5Y", "10Y", "MAX"
    ] = "1Y",
    **kargs,
) -> List[Nav]:
    """
    Gets the latest NAV

    **Parameters:**

    * **fund_name** - Fund name found in finnomena such as `TISTECH-A`
    * **source** - *(optional)* Data source for pull data. See Data Sources
    section in the documentation for all availiable options.
    * **range** - *(optional)* time period defalut to 1 year, avaliable options are "1D", "1W", "1M", "6M", "YTD", "1Y", "3Y", "5Y", "10Y", "MAX"
    * **asDataFrame** - *(optional)* return pandas dataframe instead.
    * **subscription_key** - *(optional)* Subscription key that required for
    a data source like `sec` (a.k.a)


    **Returns:** `List[Nav]` or `pd.DataFrame`

    Usage:
    ```
    >>> import pythainav as nav

    >>> nav.get_all("KT-PRECIOUS")
    [Nav(value=4.2696, updated='20/01/2020', tags={'latest'}, fund='KT-PRECIOUS'), ...]

    >>> nav.get_all("KT-PRECIOUS", asDataFrame=True)
            value    updated tags         fund
    0     10.0001 2010-11-19   {}  KT-PRECIOUS
    1     10.0566 2010-11-22   {}  KT-PRECIOUS
    2     10.0326 2010-11-23   {}  KT-PRECIOUS
    3     10.0428 2010-11-24   {}  KT-PRECIOUS
    4     10.0253 2010-11-25   {}  KT-PRECIOUS
    ...       ...        ...  ...          ...
    2260   5.5777 2020-10-07   {}  KT-PRECIOUS
    2261   5.6468 2020-10-08   {}  KT-PRECIOUS
    2262   5.8868 2020-10-09   {}  KT-PRECIOUS
    2263   5.9086 2020-10-14   {}  KT-PRECIOUS
    2264   5.8438 2020-10-15   {}  KT-PRECIOUS

    [2265 rows x 4 columns]
    ```
    """
    fund_name = fund_name

    source2class = {
        "finnomena": sources.Finnomena,
        "sec": sources.Sec,
    }
    _source = source2class[source](**kargs)

    navs = _source.get_range(fund_name, range=range)

    if asDataFrame:
        pd = import_optional_dependency("pandas")
        from dataclasses import asdict

        navs = pd.DataFrame([asdict(x) for x in navs])

    return navs



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pythainav', 'pythainav.utils']

package_data = \
{'': ['*']}

install_requires = \
['dateparser>=1.0.0,<2.0.0',
 'furl>=2.1,<3.0',
 'fuzzywuzzy[speedup]>=0.17,<0.19',
 'importlib-metadata>=4.8.1,<5.0.0',
 'requests>=2.22,<3.0',
 'typing-extensions>=3.10.0,<4.0.0']

extras_require = \
{':python_version == "3.6"': ['dataclasses>=0.7.0,<0.8.0']}

setup_kwargs = {
    'name': 'pythainav',
    'version': '0.2.1',
    'description': 'a Python interface to pull thai mutual fund NAV',
    'long_description': '# PythaiNAV: ทำให้การดึงข้อมูลกองทุนไทยเป็นเรื่องง่าย\n<!-- ALL-CONTRIBUTORS-BADGE:START - Do not remove or modify this section -->\n[![All Contributors](https://img.shields.io/badge/all_contributors-5-orange.svg?style=flat-square)](#contributors-)\n<!-- ALL-CONTRIBUTORS-BADGE:END -->\n[![FOSSA Status](https://app.fossa.io/api/projects/git%2Bgithub.com%2FCircleOnCircles%2Fpythainav.svg?type=shield)](https://app.fossa.io/projects/git%2Bgithub.com%2FCircleOnCircles%2Fpythainav?ref=badge_shield)\n![Tests](https://github.com/CircleOnCircles/pythainav/workflows/Tests/badge.svg?branch=master)\n[![codecov](https://codecov.io/gh/CircleOnCircles/pythainav/branch/develop/graph/badge.svg)](https://codecov.io/gh/CircleOnCircles/pythainav)\n[![Codacy Badge](https://api.codacy.com/project/badge/Grade/f868488db4ba4266a112c3432301c6b4)](https://www.codacy.com/manual/nutchanon/pythainav?utm_source=github.com&amp;utm_medium=referral&amp;utm_content=CircleOnCircles/pythainav&amp;utm_campaign=Badge_Grade)\n\n\n\n![cover image](https://github.com/CircleOnCircles/pythainav/raw/master/extra/pythainav.png)\n\n\n\n> อยากชวนทุกคนมาร่วมพัฒนา ติชม แนะนำ เพื่อให้ทุกคนเข้าถึงข้อมูลการง่ายขึ้น [เริ่มต้นได้ที่นี้](https://github.com/CircleOnCircles/pythainav/issues) หรือเข้ามา Chat ใน [Discord](https://discord.gg/jjuMcKZ) ได้เลย 😊\n\n📖 Documentation is here. คู่มือการใช้งานอยู่ที่นี่ <https://pythainav.nutchanon.org/>\n\n## Get Started - เริ่มต้นใช้งาน\n\n```bash\n$ pip install pythainav\n```\n\n```python\nimport pythainav as nav\n\nnav.get("KT-PRECIOUS")\n> Nav(value=4.2696, updated=\'20/01/2020\', tags={\'latest\'}, fund=\'KT-PRECIOUS\')\n\nnav.get("TISTECH-A", date="1 week ago")\n> Nav(value=12.9976, updated=\'14/01/2020\', tags={}, fund=\'TISTECH-A\')\n\nnav.get_all("TISTECH-A", range="MAX")\n> [Nav(value=12.9976, updated=\'21/01/2020\', tags={}, fund=\'TISTECH-A\'), Nav(value=12.9002, updated=\'20/01/2020\', tags={}, fund=\'TISTECH-A\'), ...]\n\nnav.get_all("KT-PRECIOUS", asDataFrame=True)\n> pd.DataFrame [2121 rows x 4 columns]\n```\n\n## Source of Data - ที่มาข้อมูล\n\nดูจาก <https://pythainav.nutchanon.org/datasource/>\n\n## Disclaimer\n\nเราไม่รับประกันความเสียหายใดๆทั้งสิ้นที่เกิดจาก แหล่งข้อมูล, library, source code,sample code, documentation, library dependencies และอื่นๆ\n\n## Contributors ✨\n\nThanks goes to these wonderful people ([emoji key](https://allcontributors.org/docs/en/emoji-key)):\n\n<!-- ALL-CONTRIBUTORS-LIST:START - Do not remove or modify this section -->\n<!-- prettier-ignore-start -->\n<!-- markdownlint-disable -->\n<table>\n  <tr>\n    <td align="center"><a href="http://nutchanon.org"><img src="https://avatars2.githubusercontent.com/u/8089231?v=4?s=100" width="100px;" alt=""/><br /><sub><b>Nutchanon Ninyawee</b></sub></a><br /><a href="https://github.com/CircleOnCircles/pythainav/commits?author=CircleOnCircles" title="Code">💻</a> <a href="#infra-CircleOnCircles" title="Infrastructure (Hosting, Build-Tools, etc)">🚇</a></td>\n    <td align="center"><a href="https://github.com/sctnightcore"><img src="https://avatars2.githubusercontent.com/u/23263315?v=4?s=100" width="100px;" alt=""/><br /><sub><b>sctnightcore</b></sub></a><br /><a href="https://github.com/CircleOnCircles/pythainav/commits?author=sctnightcore" title="Code">💻</a> <a href="#talk-sctnightcore" title="Talks">📢</a> <a href="#ideas-sctnightcore" title="Ideas, Planning, & Feedback">🤔</a></td>\n    <td align="center"><a href="https://github.com/angonyfox"><img src="https://avatars3.githubusercontent.com/u/1295513?v=4?s=100" width="100px;" alt=""/><br /><sub><b>angonyfox</b></sub></a><br /><a href="https://github.com/CircleOnCircles/pythainav/commits?author=angonyfox" title="Code">💻</a> <a href="https://github.com/CircleOnCircles/pythainav/commits?author=angonyfox" title="Tests">⚠️</a></td>\n    <td align="center"><a href="https://github.com/samupra"><img src="https://avatars.githubusercontent.com/u/24209940?v=4?s=100" width="100px;" alt=""/><br /><sub><b>Pongpira Upra</b></sub></a><br /><a href="https://github.com/CircleOnCircles/pythainav/commits?author=samupra" title="Code">💻</a></td>\n    <td align="center"><a href="https://github.com/namv2012"><img src="https://avatars.githubusercontent.com/u/50385570?v=4?s=100" width="100px;" alt=""/><br /><sub><b>namv2012</b></sub></a><br /><a href="https://github.com/CircleOnCircles/pythainav/commits?author=namv2012" title="Code">💻</a></td>\n  </tr>\n</table>\n\n<!-- markdownlint-restore -->\n<!-- prettier-ignore-end -->\n\n<!-- ALL-CONTRIBUTORS-LIST:END -->\n\nThis project follows the [all-contributors](https://github.com/all-contributors/all-contributors) specification. Contributions of any kind welcome!\n\n\n## License\n[![FOSSA Status](https://app.fossa.io/api/projects/git%2Bgithub.com%2FCircleOnCircles%2Fpythainav.svg?type=large)](https://app.fossa.io/projects/git%2Bgithub.com%2FCircleOnCircles%2Fpythainav?ref=badge_large)\n',
    'author': 'Nutchanon Ninyawee',
    'author_email': 'me@nutchanon.org',
    'maintainer': 'Nutchanon Ninyawee',
    'maintainer_email': 'me@nutchanon.org',
    'url': 'https://github.com/CircleOnCircles/pythainav',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from typing import List

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


import datetime
from abc import ABC, abstractmethod
from functools import lru_cache

import dateparser
import requests
from furl import furl

from .nav import Nav
from .utils.date import convert_buddhist_to_gregorian, date_range


class Source(ABC):
    @abstractmethod
    def get(self, fund: str):
        pass

    @abstractmethod
    def list(self):
        pass


class Finnomena(Source):
    base = furl("https://www.finnomena.com/fn3/api/fund/")
    base_v2 = furl("https://www.finnomena.com/fn3/api/fund/v2/")

    def __init__(self):
        super().__init__()

    def _find_earliest(self, navs: List[Nav], date: str):
        navs.sort(key=lambda x: x.updated, reverse=True)

        date = dateparser.parse(date)
        for nav in navs:
            if date >= nav.updated:
                return nav
        return None

    def get(self, fund: str, date: str = None):

        if date:
            navs = self.get_range(fund)
            return self._find_earliest(navs, date)

        name2fund = self.list()

        url = self.base / "nav" / "latest"
        url.args["fund"] = name2fund[fund]["id"]

        # convert to str
        url = url.url

        nav = requests.get(url).json()
        nav = Nav(
            value=float(nav["value"]),
            updated=datetime.datetime.strptime(nav["nav_date"], "%Y-%m-%d"),
            tags={"latest"},
            fund=fund,
        )
        return nav

    # cache here should be sensible since the fund is not regulary update
    # TODO: change or ttl cache with timeout = [1 hour, 1 day]
    @lru_cache(maxsize=1024)
    def get_range_v1(self, fund: str, period="SI"):
        name2fund = self.list()

        url = self.base / "nav" / "q"
        url.args["fund"] = name2fund[fund]["id"]
        url.args["range"] = period

        # convert to str
        url = url.url

        navs_response = requests.get(url).json()

        navs = []
        for nav_resp in navs_response:
            nav = Nav(
                value=float(nav_resp["value"]),
                updated=dateparser.parse(nav_resp["nav_date"]),
                tags={},
                fund=fund,
            )
            nav.amount = nav_resp["amount"]
            navs.append(nav)

        return navs

    # cache here should be sensible since the fund is not regulary update
    # TODO: change or ttl cache with timeout = [1 hour, 1 day]
    @lru_cache(maxsize=1024)
    def get_range(
        self,
        fund: str,
        range: Literal[
            "1D", "1W", "1M", "6M", "YTD", "1Y", "3Y", "5Y", "10Y", "MAX"
        ] = "1Y",
    ):
        name2fund = self.list()

        # /fn3/api/fund/v2/ public/funds/F00000IT9T/nav/q
        url = (
            self.base_v2
            / "public"
            / "funds"
            / name2fund[fund]["id"]
            / "nav"
            / "q"
        )
        url.args["range"] = range

        # convert to str
        url = url.url

        navs_response = requests.get(url).json()

        if not navs_response["status"]:
            raise Exception(f"response to {url} is invalid")

        navs = []
        for nav_resp in navs_response["data"]["navs"]:
            date = dateparser.parse(nav_resp["date"])
            date = date.replace(tzinfo=None)
            nav = Nav(
                value=float(nav_resp["value"]),
                updated=date,
                tags={},
                fund=fund,
            )
            nav.amount = nav_resp["amount"]
            navs.append(nav)

        return navs

    # cache here should be sensible since the fund is not regulary update
    # TODO: change or ttl cache with timeout = [1 hour, 1 day]
    # TODO: New API exists /fn3/api/fund/public/filter/overview
    @lru_cache(maxsize=1)
    def list(self):
        url = self.base / "public" / "list"
        url = url.url
        funds = requests.get(url).json()
        return {fund["short_code"]: fund for fund in funds}

    # def _list(self, )


class Sec(Source):
    base = furl("https://api.sec.or.th/")

    def __init__(self, subscription_key: dict = None):
        super().__init__()
        if subscription_key is None:
            # TODO: Create specific exception for this
            raise ValueError("Missing subscription key")
        if not all(
            [
                True if key in subscription_key else False
                for key in ["fundfactsheet", "funddailyinfo"]
            ]
        ):
            raise ValueError(
                "subscription_key must contain 'fundfactsheet' and 'funddailyinfo' key"
            )
        self.subscription_key = subscription_key
        self.headers = {
            "Content-Type": "application/json",
        }
        self.base_url = {
            "fundfactsheet": self.base.copy().add(
                path=["FundFactsheet", "fund"]
            ),
            "funddailyinfo": self.base.copy().add(path="FundDailyInfo"),
        }
        self.session = requests.Session()
        self.session.headers.update(self.headers)

    def __get_api_data(
        self, url, headers=None, subscription_key="fundfactsheet"
    ):
        if headers is None:
            headers = self.headers
        headers.update(
            {
                "Ocp-Apim-Subscription-Key": self.subscription_key[
                    subscription_key
                ]
            }
        )
        response = self.session.get(url, headers=headers)
        # check status code
        response.raise_for_status()
        if response.status_code == 200:
            if response.headers["content-length"] == "0":
                return None
            return response.json()
        # No content
        elif response.status_code == 204:
            return None

    def get(self, fund: str, date: str = None):
        if date:
            if isinstance(date, str):
                query_date = dateparser.parse(date).date()
            elif isinstance(date, datetime.date):
                query_date = date
            elif isinstance(date, datetime.datetime):
                query_date = date.date()
        else:
            # TODO: Upgrade to smarter https://stackoverflow.com/questions/2224742/most-recent-previous-business-day-in-python
            # PS. should i add pandas as dep? it's so largeee.
            query_date = last_bus_day = datetime.date.today()
            wk_day = datetime.date.weekday(last_bus_day)
            if wk_day > 4:  # if it's Saturday or Sunday
                last_bus_day = last_bus_day - datetime.timedelta(
                    days=wk_day - 4
                )  # then make it Friday
            query_date = last_bus_day

        if not fund:
            raise ValueError("Must specify fund")

        list_fund = self.search(fund)
        if list_fund:
            fund_info = list_fund.pop(0)
            fund_id = fund_info["proj_id"]
            nav = self.get_nav_from_fund_id(fund_id, query_date)

            if isinstance(nav, Nav):
                nav.fund = fund_info["proj_abbr_name"]
                if query_date == datetime.date.today():
                    nav.tags = {"latest"}
                return nav
            else:
                return nav
        else:
            # Fund not found
            # due to query_date is a week day that also a holiday
            return None

    def get_range(self, fund: str, period="SI"):
        list_fund = self.search(fund)
        if list_fund:
            fund_info = list_fund.pop(0)
            today = datetime.date.today()
            if period == "SI":
                if fund_info["regis_date"] != "-":
                    inception_date = dateparser.parse(
                        fund_info["regis_date"]
                    ).date()
                    data_date = date_range(inception_date, today)
                else:
                    date_date = [today]
            else:
                query_date = dateparser.parse(period).date()
                data_date = date_range(query_date, today)
            list_nav = []
            # Remove weekend
            data_date = [
                dd for dd in date_date if dd.isoweekday() not in [6, 7]
            ]
            for dd in data_date:
                nav = self.get_nav_from_fund_id(fund, dd)
                if nav:
                    list_nav.append(nav)
            return list_nav
        else:
            # Fund not found
            return None

    @lru_cache(maxsize=1024)
    def get_nav_from_fund_id(self, fund_id: str, nav_date: datetime.date):
        url = (
            self.base_url["funddailyinfo"]
            .copy()
            .add(path=[fund_id, "dailynav", nav_date.isoformat()])
            .url
        )
        headers = self.headers
        headers.update(
            {
                "Ocp-Apim-Subscription-Key": self.subscription_key[
                    "funddailyinfo"
                ]
            }
        )
        response = self.session.get(url, headers=headers)
        # check status code
        response.raise_for_status()
        if response.status_code == 200:
            if response.headers["content-length"] == "0":
                raise requests.exceptions.ConnectionError("No data received")
            result = response.json()
            # Multi class fund
            if (
                float(result["last_val"]) == 0.0
                and float(result["previous_val"]) == 0
            ):
                remark_en = result["amc_info"][0]["remark_en"]
                multi_class_nav = {
                    k.strip(): float(v)
                    for x in remark_en.split("/")
                    for k, v in [x.split("=")]
                }
                list_nav = []
                for fund_name, nav_val in multi_class_nav.items():
                    n = Nav(
                        value=float(nav_val),
                        updated=dateparser.parse(result["nav_date"]),
                        tags={},
                        fund=fund_name,
                    )
                    n.amount = result["net_asset"]
                    list_nav.append(n)
                return list_nav
            else:
                n = Nav(
                    value=float(result["last_val"]),
                    updated=dateparser.parse(result["nav_date"]),
                    tags={},
                    fund=fund_id,
                )
                n.amount = result["net_asset"]
                return n
        # No content
        elif response.status_code == 204:
            return None

    @lru_cache(maxsize=1024)
    def list(self):
        return self.search_fund(name="")

    def search(self, name: str):
        result = self.search_fund(name)
        if result is None:
            result = self.search_class_fund(name)
        return result

    @lru_cache(maxsize=1024)
    def search_fund(self, name: str):
        url = self.base_url["fundfactsheet"].url
        headers = self.headers
        headers.update(
            {
                "Ocp-Apim-Subscription-Key": self.subscription_key[
                    "fundfactsheet"
                ]
            }
        )
        response = self.session.post(url, headers=headers, json={"name": name})
        # check status code
        response.raise_for_status()
        if response.status_code == 200:
            if response.headers["content-length"] == "0":
                raise requests.exceptions.ConnectionError("No data received")
            return response.json()
        # No content
        elif response.status_code == 204:
            return None

    @lru_cache(maxsize=1024)
    def search_class_fund(self, name: str):
        url = self.base_url["fundfactsheet"].copy().add(path="class_fund").url
        headers = self.headers
        headers.update(
            {
                "Ocp-Apim-Subscription-Key": self.subscription_key[
                    "fundfactsheet"
                ]
            }
        )
        response = self.session.post(url, headers=headers, json={"name": name})
        # check status code
        response.raise_for_status()
        if response.status_code == 200:
            if response.headers["content-length"] == "0":
                raise requests.exceptions.ConnectionError("No data received")
            return response.json()
        # No content
        elif response.status_code == 204:
            return None

    def list_amc(self):
        url = self.base_url["fundfactsheet"].copy().add(path="amc").url
        result = self.__get_api_data(url)
        return result

    def list_fund_under_amc(self, amc_id):
        if amc_id is None and not amc_id:
            raise ValueError("Missing amc_id")
        url = (
            self.base_url["fundfactsheet"].copy().add(path=["amc", amc_id]).url
        )
        result = self.__get_api_data(url)
        return result

    def get_fund_factsheet_url(self, fund_id):
        if not fund_id:
            raise ValueError("Must specify fund")
        url = (
            self.base_url["fundfactsheet"]
            .copy()
            .add(path=[fund_id, "URLs"])
            .url
        )
        result = self.__get_api_data(url)
        return result

    def get_fund_ipo(self, fund_id):
        if not fund_id:
            raise ValueError("Must specify fund")
        url = (
            self.base_url["fundfactsheet"].copy().add(path=[fund_id, "IPO"]).url
        )
        result = self.__get_api_data(url)
        return result

    def get_fund_investment(self, fund_id):
        if not fund_id:
            raise ValueError("Must specify fund")
        url = (
            self.base_url["fundfactsheet"]
            .copy()
            .add(path=[fund_id, "investment"])
            .url
        )
        result = self.__get_api_data(url)
        return result

    def get_fund_project_type(self, fund_id):
        if not fund_id:
            raise ValueError("Must specify fund")
        url = (
            self.base_url["fundfactsheet"]
            .copy()
            .add(path=[fund_id, "project_type"])
            .url
        )
        result = self.__get_api_data(url)
        return result

    def get_fund_policy(self, fund_id):
        if not fund_id:
            raise ValueError("Must specify fund")
        url = (
            self.base_url["fundfactsheet"]
            .copy()
            .add(path=[fund_id, "policy"])
            .url
        )
        result = self.__get_api_data(url)
        if "investment_policy_desc" in result and len(
            result["investment_policy_desc"]
        ):
            result["investment_policy_desc"] = base64.b64decode(
                result["investment_policy_desc"]
            ).decode("utf-8")
        return result

    def get_fund_specification(self, fund_id):
        if not fund_id:
            raise ValueError("Must specify fund")
        url = (
            self.base_url["fundfactsheet"]
            .copy()
            .add(path=[fund_id, "specification"])
            .url
        )
        result = self.__get_api_data(url)
        return result

    def get_fund_feeder_fund(self, fund_id):
        if not fund_id:
            raise ValueError("Must specify fund")
        url = (
            self.base_url["fundfactsheet"]
            .copy()
            .add(path=[fund_id, "feeder_fund"])
            .url
        )
        result = self.__get_api_data(url)
        return result

    def get_fund_redemption(self, fund_id):
        if not fund_id:
            raise ValueError("Must specify fund")
        url = (
            self.base_url["fundfactsheet"]
            .copy()
            .add(path=[fund_id, "redemption"])
            .url
        )
        result = self.__get_api_data(url)
        return result

    def get_fund_suitability(self, fund_id):
        if not fund_id:
            raise ValueError("Must specify fund")
        url = (
            self.base_url["fundfactsheet"]
            .copy()
            .add(path=[fund_id, "suitability"])
            .url
        )
        result = self.__get_api_data(url)
        for key in [
            "fund_suitable_desc",
            "fund_not_suitable_desc",
            "important_notice",
            "risk_spectrum_desc",
        ]:
            if key in result:
                result[key] = base64.b64decode(result[key]).decode("utf-8")
        return result

    def get_fund_risk(self, fund_id):
        if not fund_id:
            raise ValueError("Must specify fund")
        url = (
            self.base_url["fundfactsheet"]
            .copy()
            .add(path=[fund_id, "risk"])
            .url
        )
        result = self.__get_api_data(url)
        return result

    def get_fund_asset(self, fund_id):
        if not fund_id:
            raise ValueError("Must specify fund")
        url = (
            self.base_url["fundfactsheet"]
            .copy()
            .add(path=[fund_id, "asset"])
            .url
        )
        result = self.__get_api_data(url)
        return result

    def get_fund_turnover_ratio(self, fund_id):
        if not fund_id:
            raise ValueError("Must specify fund")
        url = (
            self.base_url["fundfactsheet"]
            .copy()
            .add(path=[fund_id, "turnover_ratio"])
            .url
        )
        result = self.__get_api_data(url)
        return result

    def get_fund_return(self, fund_id):
        if not fund_id:
            raise ValueError("Must specify fund")
        url = (
            self.base_url["fundfactsheet"]
            .copy()
            .add(path=[fund_id, "return"])
            .url
        )
        result = self.__get_api_data(url)
        return result

    def get_fund_buy_and_hold(self, fund_id):
        if not fund_id:
            raise ValueError("Must specify fund")
        url = (
            self.base_url["fundfactsheet"]
            .copy()
            .add(path=[fund_id, "buy_and_hold"])
            .url
        )
        result = self.__get_api_data(url)
        return result

    def get_fund_benchmark(self, fund_id):
        if not fund_id:
            raise ValueError("Must specify fund")
        url = (
            self.base_url["fundfactsheet"]
            .copy()
            .add(path=[fund_id, "benchmark"])
            .url
        )
        result = self.__get_api_data(url)
        return result

    def get_fund_compare(self, fund_id):
        if not fund_id:
            raise ValueError("Must specify fund")
        url = (
            self.base_url["fundfactsheet"]
            .copy()
            .add(path=[fund_id, "fund_compare"])
            .url
        )
        result = self.__get_api_data(url)
        return result

    def get_class_fund(self, fund_id):
        if not fund_id:
            raise ValueError("Must specify fund")
        url = (
            self.base_url["fundfactsheet"]
            .copy()
            .add(path=[fund_id, "class_fund"])
            .url
        )
        result = self.__get_api_data(url)
        return result

    def get_fund_performance(self, fund_id):
        if not fund_id:
            raise ValueError("Must specify fund")
        url = (
            self.base_url["fundfactsheet"]
            .copy()
            .add(path=[fund_id, "performance"])
            .url
        )
        result = self.__get_api_data(url)
        return result

    def get_fund_5yearlost(self, fund_id):
        if not fund_id:
            raise ValueError("Must specify fund")
        url = (
            self.base_url["fundfactsheet"]
            .copy()
            .add(path=[fund_id, "5YearLost"])
            .url
        )
        result = self.__get_api_data(url)
        return result

    def get_fund_dividend_policy(self, fund_id):
        if not fund_id:
            raise ValueError("Must specify fund")
        url = (
            self.base_url["fundfactsheet"]
            .copy()
            .add(path=[fund_id, "dividend"])
            .url
        )
        result = self.__get_api_data(url)
        for record in result:
            if "dividend_details" in record:
                for dividend_record in record["dividend_details"]:
                    if dividend_record["book_closing_date"] != "-":
                        dividend_record["book_closing_date"] = (
                            convert_buddhist_to_gregorian(
                                dividend_record["book_closing_date"]
                            )
                            .date()
                            .isoformat()
                        )
                    if dividend_record["payment_date"] != "-":
                        dividend_record["payment_date"] = (
                            convert_buddhist_to_gregorian(
                                dividend_record["payment_date"]
                            )
                            .date()
                            .isoformat()
                        )
        return result

    def get_fund_fee(self, fund_id):
        if not fund_id:
            raise ValueError("Must specify fund")
        url = (
            self.base_url["fundfactsheet"].copy().add(path=[fund_id, "fee"]).url
        )
        result = self.__get_api_data(url)
        return result

    def get_fund_involveparty(self, fund_id):
        if not fund_id:
            raise ValueError("Must specify fund")
        url = (
            self.base_url["fundfactsheet"]
            .copy()
            .add(path=[fund_id, "InvolveParty"])
            .url
        )
        result = self.__get_api_data(url)
        for record in result:
            if "effective_date" in record and record["effective_date"] != "-":
                record["effective_date"] = (
                    convert_buddhist_to_gregorian(record["effective_date"])
                    .date()
                    .isoformat()
                )
        return result

    def get_fund_port(self, fund_id, period):
        if not fund_id:
            raise ValueError("Must specify fund")
        url = (
            self.base_url["fundfactsheet"]
            .copy()
            .add(path=[fund_id, "FundPort", period])
            .url
        )
        result = self.__get_api_data(url)
        return result

    def get_fund_full_port(self, fund_id, period):
        if not fund_id:
            raise ValueError("Must specify fund")
        url = (
            self.base_url["fundfactsheet"]
            .copy()
            .add(path=[fund_id, "FundPort", period])
            .url
        )
        result = self.__get_api_data(url)
        return result

    def get_fund_top5_port(self, fund_id, period):
        if not fund_id:
            raise ValueError("Must specify fund")
        url = (
            self.base_url["fundfactsheet"]
            .copy()
            .add(path=[fund_id, "FundTop5", period])
            .url
        )
        result = self.__get_api_data(url)
        return result

    def get_fund_dividend(self, fund_id):
        if not fund_id:
            raise ValueError("Must specify fund")
        url = (
            self.base_url["funddailyinfo"]
            .copy()
            .add(path=[fund_id, "dividend"])
            .url
        )
        result = self.__get_api_data(url, subscription_key="funddailyinfo")
        return result

    def get_amc_submit_dailyinfo(self):
        url = self.base_url["funddailyinfo"].copy().add(path=["amc"]).url
        result = self.__get_api_data(url, subscription_key="funddailyinfo")
        return result
